/* dtmf.c  -  dual tone multifrequency generator
 * $Id: dtmf.c,v 1.16 2002/04/02 09:01:44 nlaredo Exp $
 *
 *  Copyright (C)1993-2002 Nathan Laredo (laredo@gnu.org)
 *
 * This file may be freely distributed under the terms of
 * the GNU General Public Licence (GPL).
 * 
 * NOTE: This is actually now single to 128 tone generation
 *       but I didn't feel like changing the program name.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <math.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>

#define SOUND_DEVICE "/dev/dsp"
#define SAMPLELEN  1048576
#define SAMPLERATE 48000

/* artificial limit on stdin processing of 1MB */
#define STDIN_SIZE 1048576

static unsigned char stdinbuf[STDIN_SIZE];
static unsigned short soundbuf[SAMPLELEN];
static int fd, rate, s, nyquist;
static float k;
int channels = 2;

void tonegen(int argc, char **argv)
{
	float r[128], f[128], v[128], len, sample, sample2, vol;
	int i, j, size, count;
	char *decay;

	if (argc < 3) {
		fprintf(stderr, "usage: %s seconds[d] f1 vol [f2 vol] [...] "
			"[%s same args for next in sequence, vol = 0-255]\n",
			argv[0], argv[0]);
		exit(1);
	}
	len = atof(argv[1]);
	decay = strchr(argv[1], 'd');
	for (i = count = 0; i < argc - 2 && count < 127; i++) {
		f[count] = atof(argv[i + 2]);
		vol = 0.0;
		if (i + 1 < argc - 2) {
			vol = atof(argv[i + 3]);
		}
		if (vol >= 0.003 && vol <= 1.0) {
			i++;
		} else {
			vol = 1.0;
		}
		v[count] = vol;
		if ((int)f[count] < -nyquist || (int)f[count] > nyquist)
			f[count] = 0.0;
		r[count] = 2.0 * M_PI * f[count] / (float)rate;
		count++;
	}
	len *= (float)rate * channels;
	size = (int)len;
	if (size > SAMPLELEN)
		size = SAMPLELEN;	/* one second playback */
	if (size < 1)
		return;

	for (i = 0; i < (size / channels); i++) {
		sample = 0.0;
		sample2= 0.0;
		for (j = 0; j < count; j++) {
			sample += sin(r[j] * (i + s)) * v[j];
			sample2 += sin(r[j] * (i * 1.003 + s)) * v[j];
		}
		sample *= 32767.0 / (int)(count);
		sample2 *= 32767.0 / (int)(count);
		if (decay) {
			sample *= exp(k * (float)i);
			sample2 *= exp(k * (float)i);
		}
#if 0
		if (i > size - 80) {	/* taper to zero at end */
			sample *= (float)(size - i) / 200.0;
			sample2 *= (float)(size - i) / 200.0;
		}

		if (i < 80) {	/* taper from zero at start */
			sample *= (float)(i) / 200.0;
			sample2 *= (float)(i) / 200.0;
		}
#endif

		soundbuf[i * channels] = (int)sample;
		soundbuf[i * channels + 1] = (int)sample2;
	}
	s += size;	/* maintain phase relationship */
	write (fd, soundbuf, size << 1);
}

void process_buf(char *buf, int len)
{
	int argc, where = 0, count;
	char *argv[1048576];	/* waste! I'm feeling lazy though */
	
	argv[0] = strtok(buf, " \n");
	for (argc = 1; argc < 1048576; argc++) {
		argv[argc] = strtok(NULL, " \n");
		if (!argv[argc])
			break;
	}
	while (where < argc) {
		for (count = where + 1; count < argc ; count++)
			if ((strcmp(argv[0], argv[count]) == 0))
				break;
		tonegen(count - where, &argv[where]);
		where = count;
	}
}

int main(int argc, char **argv)
{
	int where = 0, count;

	if ((fd = open (SOUND_DEVICE, O_WRONLY)) < 0) {
		perror(SOUND_DEVICE);
		exit(1);
	}
	ioctl(fd, SNDCTL_DSP_CHANNELS, &channels);
	count = AFMT_S16_LE;
	rate = SAMPLERATE;
	ioctl(fd, SNDCTL_DSP_SPEED, &rate);
	count = AFMT_S16_LE;
	ioctl(fd, SNDCTL_DSP_SETFMT, &count);
	nyquist = rate >> 1;

	/* calculate exponential decay factor, 200ms to 50% */
	k = log(0.5) / ((float) rate * 0.2);

	if (argc > 1 && *argv[1] == '-') {
		/* take arguments from stdin */
		count = fread(stdinbuf, 1, STDIN_SIZE, stdin);
		process_buf(stdinbuf, count);
	} else {
		/* take arguments from command line */
		while (where < argc) {
			for (count = where + 1; count < argc ; count++)
				if ((strcmp(argv[0], argv[count]) == 0))
					break;
			tonegen(count - where, &argv[where]);
			where = count;
		}
	}
	close(fd);
	exit(0);
}
